from os import X_OK, access
from pathlib import Path

from yaml import FullLoader, dump, load

from recoverpy.utils import errors
from recoverpy.utils.logger import LOGGER
from recoverpy.utils.saver import SAVER

_CONFIG_DIR = Path(__file__).parent.absolute()


def is_path_valid(string_path: str) -> bool:
    path: Path = Path(string_path)
    try:
        if not path.is_dir():
            return False
    except PermissionError:
        return False

    return access(path, X_OK)


def write_config_to_file(save_path: str, log_path: str, enable_logging: bool):
    with open(_CONFIG_DIR / "config.yaml", "w") as config_file:
        config: dict = {
            "save_directory": str(save_path),
            "log_directory": str(log_path),
            "enable_logging": enable_logging,
        }

        dump(config, config_file)
    load_config()


def load_config():
    with open(_CONFIG_DIR / "config.yaml", "r") as config_file:
        config: dict = load(config_file, Loader=FullLoader)

    if "save_directory" not in config or config["save_directory"] == "":
        raise errors.NoSavePath
    if not is_path_valid(config["save_directory"]):
        raise errors.InvalidSavePath

    SAVER.save_path = Path(config["save_directory"])

    if "log_directory" not in config or config["log_directory"] == "":
        LOGGER.log_enabled = False
    elif not is_path_valid(config["log_directory"]):
        raise errors.InvalidLogPath
    else:
        LOGGER.log_path = Path(config["log_directory"])
        LOGGER.log_enabled = bool(config["enable_logging"])

    if LOGGER.log_enabled:
        LOGGER.start_logging()



from logging import getLogger

from recoverpy.config.config import load_config
from recoverpy.config.setup import setup
from recoverpy.ui.handler import SCREENS_HANDLER

getLogger(__name__)


def main():
    """Set configuration and start UI."""
    setup()
    load_config()
    SCREENS_HANDLER.open_screen("parameters")



class NoSavePath(Exception):
    def __init__(self):
        super().__init__(
            "No path entered for saving in config file. Verify your 'config.yaml' file."
        )


class InvalidSavePath(Exception):
    def __init__(self):
        super().__init__(
            "Path entered for saving in config file is invalid. "
            "Verify your 'config.yaml' file."
        )


class InvalidLogPath(Exception):
    def __init__(self):
        super().__init__(
            "Path entered for logging in config file is invalid. "
            "Verify your 'config.yaml' file\n"
            'You can leave the location blank ( "" ) if you want to disable it.'
        )



from py_cui import PyCUI

from recoverpy.ui.contents import init_ui


class Screen:
    def __init__(self, master: PyCUI):
        self.master = master
        self.master.set_refresh_timeout(1)

    def create_ui_content(self):
        init_ui(self)



from py_cui.widgets import Button, ScrollTextBlock

from recoverpy.ui import handler


def set(screen):
    screen.previous_button: Button = screen.master.add_button(
        "<",
        3,
        0,
        row_span=3,
        column_span=1,
        padx=1,
        pady=0,
        command=screen.display_previous_block,
    )
    screen.previous_button.set_color(1)

    screen.next_button: Button = screen.master.add_button(
        ">",
        3,
        9,
        row_span=3,
        column_span=1,
        padx=1,
        pady=0,
        command=screen.display_next_block,
    )
    screen.next_button.set_color(1)

    screen.blockcontent_box: ScrollTextBlock = screen.master.add_text_block(
        "Block content:", 0, 1, row_span=9, column_span=8, padx=1, pady=0
    )
    screen.blockcontent_box.set_title(f"Block {screen.current_block}")

    screen.add_blockbutton: Button = screen.master.add_button(
        "Add current block to file",
        9,
        0,
        row_span=1,
        column_span=5,
        padx=1,
        pady=0,
        command=screen.add_block_to_file,
    )
    screen.add_blockbutton.set_color(6)

    screen.save_file_button: Button = screen.master.add_button(
        "Save file",
        9,
        5,
        row_span=1,
        column_span=3,
        padx=1,
        pady=0,
        command=screen.save_file,
    )
    screen.save_file_button.set_color(4)

    screen.go_back_button: Button = screen.master.add_button(
        "Go back to previous screen",
        9,
        8,
        row_span=1,
        column_span=2,
        padx=1,
        pady=0,
        command=handler.SCREENS_HANDLER.go_back,
    )
    screen.go_back_button.set_color(2)



from io import BufferedReader
from queue import Queue
from re import findall
from subprocess import DEVNULL, PIPE, Popen, check_output
from threading import Thread
from time import sleep

from recoverpy.ui.screen import Screen
from recoverpy.utils.helper import is_dependency_installed
from recoverpy.utils.logger import LOGGER


class SearchEngine:
    def start_search(self, search_screen: Screen):
        grep_process: Popen = self.create_grep_process(
            searched_string=search_screen.searched_string,
            partition=search_screen.partition,
        )

        self.start_progress_monitoring_thread(grep_process, search_screen)
        self.start_result_enqueue_thread(grep_process, search_screen)
        self.start_result_dequeue_thread(search_screen)

    def create_grep_process(self, searched_string: str, partition: str) -> Popen:
        return Popen(
            ["grep", "-a", "-b", searched_string, partition],
            stdin=None,
            stdout=PIPE,
            stderr=None,
        )

    def start_progress_monitoring_thread(
        self, grep_process: Popen, search_screen: Screen
    ):
        if not is_dependency_installed(command="progress"):
            return

        Thread(
            target=self.monitor_search_progress,
            args=(search_screen, grep_process.pid),
            daemon=True,
        ).start()
        LOGGER.write("debug", "Started progress monitoring thread")

    def start_result_enqueue_thread(self, grep_process: Popen, search_screen: Screen):
        Thread(
            target=self.enqueue_grep_output,
            args=(grep_process.stdout, search_screen.queue_object),
            daemon=True,
        ).start()
        LOGGER.write("debug", "Started grep searching thread")

    def start_result_dequeue_thread(self, search_screen: Screen):
        Thread(
            target=search_screen.dequeue_results,
            daemon=True,
        ).start()
        LOGGER.write("debug", "Started grep output fetching thread")

    def monitor_search_progress(self, search_screen: Screen, grep_pid: int):
        while True:
            output: str = check_output(
                ["progress", "-p", str(grep_pid)], stderr=DEVNULL
            ).decode("utf8")

            if not output:
                search_screen.set_title("100% - Search completed")
                return

            progress: list = findall(r"([0-9]+\.[0-9]+\%[^\)]+\))", output)
            if not progress:
                continue

            LOGGER.write("debug", f"Progress: {progress[0]}")
            search_screen.set_title(progress[0])
            sleep(1)

    def enqueue_grep_output(self, out: BufferedReader, queue: Queue):
        for line in iter(out.readline, b""):
            queue.put(line)
        out.close()

    def get_new_results(self, queue_object: Queue, current_blockindex: int) -> tuple:
        if len(list(queue_object.queue)) == current_blockindex:
            return tuple()

        queue_list: list = list(queue_object.queue)
        new_blockindex: int = len(queue_list)
        new_results: list = queue_list[current_blockindex:]

        return new_results, new_blockindex

    @staticmethod
    def get_dd_output(partition: str, block_size: int, block_number: str) -> bytes:
        return check_output(
            [
                "dd",
                f"if={partition}",
                "count=1",
                "status=none",
                f"bs={block_size}",
                f"skip={block_number}",
            ]
        )


SEARCH_ENGINE: SearchEngine = SearchEngine()



from subprocess import CalledProcessError
from typing import Optional

from py_cui import PyCUI

from recoverpy.ui.screen import Screen
from recoverpy.utils.helper import decode_result, get_block_size
from recoverpy.utils.logger import LOGGER
from recoverpy.utils.search import SearchEngine


class MenuWithBlockDisplay(Screen):
    """Composition aimed class for menus needing methods to display partition
    blocks content.
    """

    def __init__(self, master: PyCUI):
        super().__init__(master)

        self.horizontal_char_limit: int = 0

        self.current_block: Optional[str] = None
        self.current_result: Optional[str] = None
        self.partition: Optional[str] = None

    def get_dd_result(self, block_number: str = None):
        if block_number is None:
            block_number = self.current_block

        LOGGER.write(
            "debug",
            f"Getting 'dd' output for block {str(self.current_block)}",
        )

        try:
            dd_result: bytes = SearchEngine.get_dd_output(
                partition=self.partition,
                block_size=get_block_size(self.partition),
                block_number=block_number,
            )

            self.current_result = decode_result(dd_result)
            self.current_block = block_number

            LOGGER.write("debug", "dd command successful")
        except CalledProcessError:
            self.master.show_error_popup(
                "Mmmmhhh...",
                f"Error while opening block {str(self.current_block)}",
            )
            LOGGER.write(
                "error",
                f"Error while opening block {str(self.current_block)}",
            )

    def update_textbox(self):
        self.update_horizontal_char_limit()

        # Format raw result to fit in the text box
        blocklines: list = [
            str(self.current_result)[i : i + self.horizontal_char_limit]
            for i in range(
                0,
                len(str(self.current_result)) + self.horizontal_char_limit,
                self.horizontal_char_limit,
            )
        ]
        formated_result: str = "\n".join(blocklines)

        # Fix for embedded null character
        formated_result = formated_result.replace(chr(0), "")

        self.blockcontent_box.set_text(formated_result)
        self.blockcontent_box.set_title(f"Block {self.current_block}")

        LOGGER.write("debug", f"Textbox updated with block {self.current_block}")

    def display_previous_block(self):
        try:
            self.display_block(str(int(self.current_block) - 1))
        except ValueError:
            LOGGER.write("error", f"Cannot display block {self.current_block} - 1")
            return

    def display_next_block(self):
        try:
            self.display_block(str(int(self.current_block) + 1))
        except ValueError:
            LOGGER.write("error", f"Cannot display block {self.current_block} + 1")
            return

    def display_block(self, block_number: str):
        if int(block_number) < 0:
            return

        self.get_dd_result(block_number=block_number)
        self.update_textbox()

    def update_horizontal_char_limit(self):
        text_box_dimensions: tuple = (
            self.blockcontent_box.get_cursor_limits_horizontal()
        )
        self.horizontal_char_limit: int = (
            text_box_dimensions[1] - text_box_dimensions[0]
        )
        LOGGER.write(
            "debug",
            f"Textbox char limit set to {self.horizontal_char_limit}",
        )



from datetime import datetime
from pathlib import Path

from recoverpy.utils.logger import LOGGER


class Saver:
    """Encapsulates all result saving related methods."""

    def __init__(self):
        self.save_path: Path = None
        self.last_saved_file: Path = None

    def save_result_string(self, result: str):
        time_format: str = datetime.now().strftime("recoverpy-save-%Y-%m-%d-%H%M%S")
        file_name: Path = self.save_path / time_format

        self.write_to_file(file_name=file_name, content=result)

    def save_result_dict(self, results: dict):
        ordered_blocks: list = [results[num] for num in sorted(results.keys())]
        final_output: str = "\n".join(ordered_blocks)

        self.save_result_string(final_output)

    def write_to_file(self, file_name: Path, content: str):
        with open(file_name, "w") as save_file:
            save_file.write(content)

        self.last_saved_file = file_name

        LOGGER.write("info", f"Output saved in file {file_name}")


SAVER: Saver = Saver()



from datetime import datetime
from logging import DEBUG, FileHandler, Logger, getLogger
from pathlib import Path


class CustomLogger:
    """Logging wrapper object."""

    def __init__(self):
        self.log_path: Path = None
        self.log_enabled: bool = False
        self._logger = None

    def start_logging(self):
        time = datetime.now().strftime("%Y-%m-%d-%H%M%S")
        log_file_name: Path = self.log_path / f"recoverpy-{time}.log"

        self._logger: Logger = getLogger("main")
        self._logger.setLevel(DEBUG)

        file_handler = FileHandler(log_file_name)
        file_handler.setLevel(DEBUG)

        self._logger.addHandler(file_handler)

    def write(self, log_type: str, text: str):
        if not self.log_enabled:
            return

        if log_type == "debug":
            self._logger.debug(text)
        elif log_type == "info":
            self._logger.info(text)
        elif log_type == "warning":
            self._logger.warning(text)
        elif log_type == "error":
            self._logger.error(text)
        elif log_type == "critical":
            self._logger.critical(text)


LOGGER = CustomLogger()



from re import findall

from py_cui import PyCUI

from recoverpy.ui import handler
from recoverpy.ui.screen import Screen
from recoverpy.utils import helper
from recoverpy.utils.logger import LOGGER


class ParametersScreen(Screen):
    """Select a partion and type a string to search."""

    def __init__(self, master: PyCUI):
        super().__init__(master)

        self.partition_to_search: str
        self.string_to_search: str
        self.partitions_dict: dict

        helper.is_user_root(window=self.master)
        self.create_ui_content()
        self.get_system_partitions()

    def get_system_partitions(self):
        self.partitions_dict = helper.get_partitions()
        if not self.partitions_dict:
            LOGGER.write("Error", "No partition found !")
            self.master.show_error_popup("Hum...", "No partition found.")
            return

        self.add_partitions_to_list()

    def add_partitions_to_list(self):
        if self.partitions_dict is None:
            return

        for partition in self.partitions_dict:
            if self.partitions_dict[partition]["IS_MOUNTED"]:
                self.partitions_list_scroll_menu.add_item(
                    f"Name: {partition}  -  "
                    f"Type: {self.partitions_dict[partition]['FSTYPE']}  -  "
                    f"Mounted at: {self.partitions_dict[partition]['MOUNT_POINT']}"
                )
            else:
                self.partitions_list_scroll_menu.add_item(
                    f"Name: {partition}  -  "
                    f"Type: {self.partitions_dict[partition]['FSTYPE']}"
                )

            LOGGER.write("debug", f"Partition added to list: {partition}")

    def select_partition(self):
        selected_partition = findall(
            r"Name\:\ ([^\ \n]+)\ ",
            self.partitions_list_scroll_menu.get(),
        )[0]

        if self.partitions_dict[selected_partition]["IS_MOUNTED"]:
            # Warn the user to unmount his partition before searching in it
            self.master.show_warning_popup(
                "You probably should unmount first !",
                f"It is highly recommended to unmount {selected_partition}"
                " ASAP to avoid any data loss.",
            )
        else:
            self.master.show_message_popup(
                "",
                f"Partition {selected_partition} selected.",
            )

        self.partition_to_search = f"/dev/{selected_partition.strip()}"

        LOGGER.write(
            "info",
            f"Partition selected: {self.partition_to_search}",
        )

    def confirm_search(self):
        if not helper.is_user_root(window=self.master):
            return

        self.string_to_search = self.string_text_box.get()

        if self.partition_to_search == "":
            # No partition selected
            self.master.show_message_popup(
                "Whoops !",
                "You have to select a partition to search.",
            )
            LOGGER.write("warning", "No partition selected for search")
        elif not self.string_to_search.strip():
            # Blank string to search
            self.master.show_message_popup(
                "Oops !",
                "You have to enter a text to search.",
            )
            LOGGER.write("warning", "No string given for search")
        else:
            # Prompt to confirm string
            self.master.show_yes_no_popup(
                "Do you want to search this text on partition "
                f"{self.partition_to_search} ?",
                self.start_search,
            )

    def start_search(self, is_confirmed: bool):
        if is_confirmed:
            LOGGER.write("info", "Starting search")
            handler.SCREENS_HANDLER.open_screen(
                "search",
                partition=self.partition_to_search,
                string_to_search=self.string_to_search.strip(),
            )



from sys import exit
from typing import Dict, Type

from py_cui import PyCUI

from recoverpy.ui import (
    screen,
    screen_block,
    screen_config,
    screen_parameters,
    screen_search,
)


class ScreensHandler:
    """Provide ui navigation logic."""

    SCREENS_CLASSES: Dict[str, Type] = {
        "parameters": screen_parameters.ParametersScreen,
        "config": screen_config.ConfigScreen,
        "search": screen_search.SearchScreen,
        "block": screen_block.BlockScreen,
    }

    def __init__(self):
        self.screens: Dict[str, screen.Screen] = {}
        self.init_screens()
        self.current_screen: str = None
        self.previous_screen: str = None

    def init_screens(self):
        for screen_class in self.SCREENS_CLASSES:
            self.screens[screen_class] = None

    def create_py_cui_master(self) -> PyCUI:
        master: PyCUI = PyCUI(10, 10)
        master.toggle_unicode_borders()
        master.set_title("RecoverPy 1.5.0")
        master.run_on_exit(exit)

        return master

    def open_screen(self, screen_name: str, **kwargs):
        self.close_screen(self.current_screen)

        master = self.create_py_cui_master()
        created_screen_object = self.SCREENS_CLASSES[screen_name](master, **kwargs)
        self.screens[screen_name] = created_screen_object

        self.current_screen, self.previous_screen = (
            screen_name,
            self.current_screen,
        )
        created_screen_object.master.start()

    def close_screen(self, screen_name: str):
        if screen_name is None or self.screens[screen_name] is None:
            return
        self.screens[screen_name].master.stop()

    def go_back(self):
        self.close_screen(self.current_screen)
        self.current_screen, self.previous_screen = (
            self.previous_screen,
            self.current_screen,
        )
        self.screens[self.current_screen].master.start()


SCREENS_HANDLER: ScreensHandler = ScreensHandler()



from queue import Queue
from re import findall
from shlex import quote
from time import sleep

from py_cui import PyCUI

from recoverpy.ui import handler
from recoverpy.ui.screen_with_block_display import MenuWithBlockDisplay
from recoverpy.utils.helper import decode_printable, get_block_size
from recoverpy.utils.logger import LOGGER
from recoverpy.utils.saver import SAVER
from recoverpy.utils.search import SEARCH_ENGINE


class SearchScreen(MenuWithBlockDisplay):
    """Display search results and corresponding blocks content."""

    def __init__(self, master: PyCUI, partition: str, string_to_search: str):
        super().__init__(master)

        self.queue_object: Queue = Queue()
        self.blockindex: int = 0
        self.block_numbers: list = []
        self.partition: str = partition
        self.block_size: int = get_block_size(partition)
        self.searched_string: str = string_to_search
        self._encoded_search_string: bytes = string_to_search.encode()

        self.create_ui_content()

        SEARCH_ENGINE.start_search(self)
        LOGGER.write(
            "info",
            f"Raw searched string:\n{self.searched_string}\n"
            f"Formated searched string:\n{quote(self.searched_string)}",
        )

    def set_title(self, grep_progress: str = None):
        title: str = (
            f"{grep_progress} - {self.blockindex} results"
            if grep_progress
            else f"{self.blockindex} results"
        )

        self.master.set_title(title)

        if "100%" in title:
            if self.blockindex == 0:
                self.master.title_bar.set_color(22)
            else:
                self.master.title_bar.set_color(30)

    def dequeue_results(self):
        while True:
            output: tuple = SEARCH_ENGINE.get_new_results(
                self.queue_object, self.blockindex
            )
            if not output:
                sleep(1)
                continue
            new_results: list = output[0]
            self.blockindex = output[1]

            self.add_results_to_list(new_results=new_results)
            self.set_title()

            # Sleep to avoid unnecessary overload
            sleep(1)

    def add_results_to_list(self, new_results: list):
        for result in new_results:
            string_result: str = decode_printable(result)
            inode: str = findall(r"^([0-9]+)\:", string_result)[0]
            result_block_offset = self.get_result_block_offset(result)
            real_result_block_start: int = (
                int(int(inode) / self.block_size) + result_block_offset
            )
            self.block_numbers.append(str(real_result_block_start))

            content_start: int = self.get_content_start(inode, string_result)
            content: str = string_result[content_start:]
            self.search_results_scroll_menu.add_item(content)

    def get_result_block_offset(self, result: bytes) -> int:
        result_index: int = result.index(self._encoded_search_string)
        return int(result_index / self.block_size)

    def get_content_start(self, inode: str, result: str) -> int:
        searched_string_pos: int = result.index(self.searched_string)

        box_start_pos: int = self.search_results_scroll_menu.get_absolute_start_pos()[0]
        box_stop_pos: int = self.search_results_scroll_menu.get_absolute_stop_pos()[0]
        box_length: int = box_stop_pos - box_start_pos

        is_result_outside_box: bool = (
            searched_string_pos + len(self.searched_string) > box_length
        )
        is_result_longer_than_box: bool = len(result) - searched_string_pos > box_length

        if is_result_outside_box and is_result_longer_than_box:
            return searched_string_pos
        elif is_result_outside_box:
            return len(result) - box_length
        else:
            return len(inode) + 1

    def update_block_number(self):
        self.current_block = self.block_numbers[
            int(self.search_results_scroll_menu.get_selected_item_index())
        ]

        LOGGER.write("debug", f"Displayed block set to {self.current_block}")

    def display_selected_block(self):
        self.update_block_number()
        self.display_block(self.current_block)

    def open_save_popup(self):
        if self.current_block is None:
            self.master.show_message_popup(
                "",
                "Please select a block first.",
            )
            return

        screen_choices: list = [
            "Save currently displayed block",
            "Explore neighboring blocks and save it all",
            "Cancel",
        ]
        self.master.show_menu_popup(
            "How do you want to save it ?",
            screen_choices,
            self.handle_save_popup_choice,
        )

    def handle_save_popup_choice(self, choice: str):
        if choice == "Explore neighboring blocks and save it all":
            handler.SCREENS_HANDLER.open_screen(
                "block",
                partition=self.partition,
                initial_block=self.current_block,
            )
        elif choice == "Save currently displayed block":
            SAVER.save_result_string(result=self.current_result)
            self.master.show_message_popup("", "Result saved.")



from os import environ

from recoverpy.utils.helper import is_dependency_installed


def verify_terminal_conf():
    """Fix for older terminals."""
    term: str = environ.get("TERM")

    if term != "xterm-256color":
        environ["TERM"] = "xterm-256color"


def verify_dependencies():
    needed_dependencies: list = ["grep", "dd", "lsblk"]

    for dep in needed_dependencies:
        if not is_dependency_installed(command=dep):
            raise OSError(
                f"Command '{dep}' is unavailable on your system.\n"
                "Please verify your configuration."
            )


def setup():
    verify_dependencies()
    verify_terminal_conf()



from py_cui import BLACK_ON_GREEN, keys
from py_cui.widgets import Button, ScrollMenu, ScrollTextBlock


def set(screen):
    screen.search_results_scroll_menu: ScrollMenu = screen.master.add_scroll_menu(
        "Search results:", 0, 0, row_span=10, column_span=5, padx=1, pady=0
    )
    screen.search_results_scroll_menu.add_text_color_rule(
        screen.searched_string,
        BLACK_ON_GREEN,
        "contains",
        match_type="regex",
    )
    screen.search_results_scroll_menu.add_key_command(
        keys.KEY_ENTER,
        screen.display_selected_block,
    )

    screen.blockcontent_box: ScrollTextBlock = screen.master.add_text_block(
        "Block content:", 0, 5, row_span=9, column_span=5, padx=1, pady=0
    )
    screen.blockcontent_box.add_key_command(
        keys.KEY_F5,
        screen.open_save_popup,
    )
    screen.blockcontent_box.add_key_command(
        keys.KEY_F6,
        screen.display_previous_block,
    )
    screen.blockcontent_box.add_key_command(
        keys.KEY_F7,
        screen.display_next_block,
    )
    screen.blockcontent_box.add_text_color_rule(
        screen.searched_string,
        BLACK_ON_GREEN,
        "contains",
        match_type="regex",
    )

    screen.previous_button: Button = screen.master.add_button(
        "<",
        9,
        5,
        row_span=1,
        column_span=1,
        padx=1,
        pady=0,
        command=screen.display_previous_block,
    )
    screen.previous_button.set_color(1)

    screen.next_button: Button = screen.master.add_button(
        ">",
        9,
        8,
        row_span=1,
        column_span=1,
        padx=1,
        pady=0,
        command=screen.display_next_block,
    )
    screen.next_button.set_color(1)

    screen.save_file_button: Button = screen.master.add_button(
        "Save Block",
        9,
        6,
        row_span=1,
        column_span=2,
        padx=1,
        pady=0,
        command=screen.open_save_popup,
    )
    screen.save_file_button.set_color(4)

    screen.exit_button: Button = screen.master.add_button(
        "Exit",
        9,
        9,
        row_span=1,
        column_span=1,
        padx=1,
        pady=0,
        command=screen.master.stop,
    )
    screen.exit_button.set_color(3)



from py_cui import PyCUI

from recoverpy.ui.screen_with_block_display import MenuWithBlockDisplay
from recoverpy.utils.logger import LOGGER as LOGGER
from recoverpy.utils.saver import SAVER as SAVER


class BlockScreen(MenuWithBlockDisplay):
    """Navigate through found partition blocks and save results."""

    def __init__(self, master: PyCUI, partition: str, initial_block: int):
        super().__init__(master)

        self.partition: str = partition
        self.saved_blocks_dict: dict = {}
        self.current_block: int = initial_block

        self.create_ui_content()
        self.display_block(self.current_block)

    def add_block_to_file(self):
        if self.current_block in self.saved_blocks_dict:
            return

        self.master.show_message_popup("", "Result added to file")
        self.saved_blocks_dict[self.current_block] = self.current_result
        LOGGER.write(
            "debug",
            f"Stored block {self.current_block} for future save",
        )

    def save_file(self):
        len_results: int = len(self.saved_blocks_dict)

        if len_results == 0:
            self.master.show_message_popup("", "No result to save yet")
            return

        SAVER.save_result_dict(self.saved_blocks_dict)

        if len_results == 1:
            self.master.show_message_popup(
                "",
                f"Block saved in {SAVER.last_saved_file}",
            )
        else:
            self.master.show_message_popup(
                "",
                f"{len_results} blocks saved in {SAVER.last_saved_file}",
            )



from py_cui.widgets import Button, ScrollTextBlock

from recoverpy.ui import handler
from recoverpy.utils.logger import LOGGER
from recoverpy.utils.saver import SAVER


def set(screen):
    screen.save_path_box: ScrollTextBlock = screen.master.add_text_box(
        title="Save Path",
        row=0,
        column=1,
        row_span=1,
        column_span=8,
        padx=0,
        pady=0,
        initial_text=str(SAVER.save_path),
    )

    screen.master.add_button(
        "Save",
        row=1,
        column=8,
        row_span=1,
        column_span=1,
        padx=0,
        pady=0,
        command=screen.set_save_path,
    ).set_color(1)

    screen.log_path_box: ScrollTextBlock = screen.master.add_text_box(
        title="Log Path",
        row=2,
        column=1,
        row_span=1,
        column_span=8,
        padx=0,
        pady=0,
        initial_text=str(LOGGER.log_path),
    )

    screen.master.add_button(
        "Save",
        row=3,
        column=8,
        row_span=1,
        column_span=1,
        padx=0,
        pady=0,
        command=screen.set_log_path,
    ).set_color(1)

    screen.master.add_label(
        title="Enable Logging",
        row=4,
        column=4,
        row_span=1,
        column_span=2,
        padx=0,
        pady=0,
    ).selectable = False

    screen.yes_button: Button = screen.master.add_button(
        "Yes",
        row=5,
        column=3,
        row_span=1,
        column_span=1,
        padx=0,
        pady=0,
        command=screen.enable_logging,
    )

    screen.no_button: Button = screen.master.add_button(
        "No",
        row=5,
        column=6,
        row_span=1,
        column_span=1,
        padx=0,
        pady=0,
        command=screen.disable_logging,
    )

    screen.master.add_button(
        "Save & Exit",
        row=8,
        column=2,
        row_span=1,
        column_span=2,
        padx=0,
        pady=0,
        command=screen.save_all,
    ).set_color(4)

    screen.master.add_button(
        "Cancel",
        row=8,
        column=6,
        row_span=1,
        column_span=2,
        padx=0,
        pady=0,
        command=handler.SCREENS_HANDLER.go_back,
    ).set_color(2)



from py_cui import GREEN_ON_BLACK, YELLOW_ON_BLACK, keys
from py_cui.widgets import Button, ScrollMenu, ScrollTextBlock

from recoverpy.ui import handler


def set(screen):
    screen.partitions_list_scroll_menu: ScrollMenu = screen.master.add_scroll_menu(
        "Select a partition to search:", 0, 0, row_span=9, column_span=5
    )
    screen.partitions_list_scroll_menu.add_key_command(
        keys.KEY_ENTER, screen.select_partition
    )
    screen.partitions_list_scroll_menu.add_text_color_rule(
        "Mounted at",
        YELLOW_ON_BLACK,
        "contains",
    )
    screen.partitions_list_scroll_menu.set_selected_color(GREEN_ON_BLACK)

    screen.string_text_box: ScrollTextBlock = screen.master.add_text_block(
        "Enter a text to search:",
        0,
        5,
        row_span=9,
        column_span=5,
    )

    screen.confirm_search_button: Button = screen.master.add_button(
        "Start",
        9,
        4,
        row_span=1,
        column_span=2,
        padx=0,
        pady=0,
        command=screen.confirm_search,
    )
    screen.confirm_search_button.set_color(4)

    screen.open_config_button: Button = screen.master.add_button(
        "Settings",
        9,
        8,
        row_span=1,
        column_span=2,
        padx=1,
        pady=0,
        command=lambda: handler.SCREENS_HANDLER.open_screen("config"),
    )
    screen.open_config_button.set_color(1)



from py_cui import PyCUI

from recoverpy.config import config as CONFIG
from recoverpy.ui import handler as handler
from recoverpy.ui.screen import Screen
from recoverpy.utils.logger import LOGGER
from recoverpy.utils.saver import SAVER


class ConfigScreen(Screen):
    """Display and edit RecoverPy configuration."""

    def __init__(self, master: PyCUI):
        super().__init__(master)
        self._log_enabled: bool = LOGGER.log_enabled
        self.create_ui_content()
        self.set_yes_no_colors()

    def set_yes_no_colors(self):
        if self._log_enabled:
            self.yes_button.set_color(4)
            self.no_button.set_color(1)
        else:
            self.yes_button.set_color(1)
            self.no_button.set_color(4)

    def set_save_path(self):
        user_input: str = self.save_path_box.get()
        if not CONFIG.is_path_valid(user_input):
            self.master.show_error_popup("Path invalid", "Given save path is invalid.")
            return

        CONFIG.write_config_to_file(
            save_path=user_input,
            log_path=LOGGER.log_path,
            enable_logging=LOGGER.log_enabled,
        )

        self.master.show_message_popup("", "Save path changed successfully")

    def set_log_path(self):
        user_input: str = self.log_path_box.get()
        if not CONFIG.is_path_valid(user_input):
            self.master.show_error_popup("Path invalid", "Given log path is invalid.")
            return

        CONFIG.write_config_to_file(
            save_path=SAVER.save_path,
            log_path=user_input,
            enable_logging=LOGGER.log_enabled,
        )

        self.master.show_message_popup("", "Log path changed successfully")

    def enable_logging(self):
        if self._log_enabled:
            return

        self.set_log_state(enabled=True)
        self.master.show_message_popup("", "Logging enabled")

    def disable_logging(self):
        if not self._log_enabled:
            return

        self.set_log_state(enabled=False)
        self.master.show_message_popup("", "Logging disabled")

    def set_log_state(self, enabled: bool):
        self._log_enabled = enabled
        self.set_yes_no_colors()

    def save_all(self):
        save_path: str = self.save_path_box.get()
        log_path: str = self.log_path_box.get()
        if not CONFIG.is_path_valid(save_path):
            self.master.show_error_popup("Path invalid", "Given save path is invalid.")
            return
        if not CONFIG.is_path_valid(log_path):
            self.master.show_error_popup("Path invalid", "Given Log path is invalid.")
            return

        CONFIG.write_config_to_file(
            save_path=save_path,
            log_path=log_path,
            enable_logging=self._log_enabled,
        )
        handler.SCREENS_HANDLER.go_back()



from types import ModuleType
from typing import Dict

from recoverpy.ui.contents import (
    screen_block_content,
    screen_config_content,
    screen_parameters_content,
    screen_search_content,
)

SCREEN_TYPE_MAP: Dict[str, ModuleType] = {
    "ParametersScreen": screen_parameters_content,
    "ConfigScreen": screen_config_content,
    "SearchScreen": screen_search_content,
    "BlockScreen": screen_block_content,
}


def init_ui(screen):
    for screen_type in SCREEN_TYPE_MAP:
        if screen_type in str(type(screen)):
            SCREEN_TYPE_MAP[screen_type].set(screen)
            return
    raise ValueError(f"Unknown screen type: {type(screen)}")



from os import geteuid
from subprocess import DEVNULL, call, check_output

from py_cui import PyCUI

from recoverpy.utils.logger import LOGGER

_IGNORED_PARTITIONS_TYPES: tuple = (" loop ", "swap")


def is_user_root(window: PyCUI) -> bool:
    if geteuid() == 0:
        LOGGER.write("info", "User is root")
        return True

    window.show_error_popup("Not root :(", "You have to be root or use sudo.")
    LOGGER.write("warning", "User is not root")
    return False


def get_partitions() -> dict:
    lsblk_output: str = lsblk()
    return format_lsblk_output(lsblk_output)


def lsblk() -> str:
    return check_output(
        ["lsblk", "-r", "-n", "-o", "NAME,TYPE,FSTYPE,MOUNTPOINT"],
        encoding="utf-8",
    )


def format_lsblk_output(lsblk_output: str) -> dict:
    partitions_dict = {}

    for line in lsblk_output.splitlines():
        if any(word in line for word in _IGNORED_PARTITIONS_TYPES):
            continue

        values = line.strip().split()

        if len(values) < 3:
            # Ignore if no FSTYPE detected
            continue

        if len(values) < 4:
            is_mounted = False
            mount_point = None
        else:
            is_mounted = True
            mount_point = values[3]

        partitions_dict[values[0]] = {
            "FSTYPE": values[2],
            "IS_MOUNTED": is_mounted,
            "MOUNT_POINT": mount_point,
        }

    return partitions_dict


def is_dependency_installed(command: str) -> bool:
    return call(["which", command], stdout=DEVNULL) == 0


def decode_result(result: bytes) -> str:
    return result.decode("utf-8", errors="ignore")


def decode_printable(result: bytes) -> str:
    return "".join(([c for c in decode_result(result) if c.isprintable()]))


def get_block_size(partition: str) -> int:
    return int(
        check_output(
            ["blockdev", "--getbsz", partition],
            encoding="utf-8",
        )
    )
